from typing import Any, Callable, Optional
from dataclasses import dataclass


@dataclass
class ModelPipelineOutput:
    data: Optional[Any] = None
    processed_data: Optional[Any] = None
    model: Optional[Any] = None


@dataclass
class EvaluationPipelineOutput:
    data: Optional[Any] = None
    processed_data: Optional[Any] = None
    evaluation_output: Optional[Any] = None


@dataclass
class PipelineOutput:
    model_pipeline_output: Optional[ModelPipelineOutput] = None
    evaluation_pipeline_output: Optional[EvaluationPipelineOutput] = None


@dataclass
class ModelPipelineInfo:
    model_output_path: str
    data_path: str
    read_data: Callable
    pre_process: Callable
    train: Callable
    save_model: Callable[[Any, str], None]
    keep_data: bool = True
    keep_processed_data: bool = True
    keep_model: bool = True


@dataclass
class EvaluationPipelineInfo:
    model_input_path: str
    data_path: str
    read_data: Callable
    load_model: Callable[[str], Any]
    pre_process: Callable[[PipelineOutput, Any], Any]
    evaluate: Callable[[Any], Any]
    keep_data: bool = True
    keep_processed_data: bool = True
    keep_evaluation: bool = True


@dataclass
class PipelineInfo:
    model_pipeline_info: ModelPipelineInfo
    evaluation_pipeline_info: EvaluationPipelineInfo



from .data_class import (ModelPipelineOutput,
                         ModelPipelineInfo,
                         EvaluationPipelineOutput,
                         EvaluationPipelineInfo,
                         PipelineOutput,
                         PipelineInfo)


def execute_model_pipeline(mpi: ModelPipelineInfo) -> ModelPipelineOutput:
    mpo = ModelPipelineOutput()
    data = mpi.read_data(mpi.data_path)
    if mpi.keep_data:
        mpo.data = data
    processed_data = mpi.pre_process(data)
    if mpi.keep_processed_data:
        mpo.processed_data = processed_data
    model = mpi.train(processed_data)
    mpi.save_model(model, mpi.model_output_path)
    if mpi.keep_model:
        mpo.model = model
    return mpo


def execute_evaluation_pipeline(epi: EvaluationPipelineInfo) -> EvaluationPipelineOutput:
    epo = EvaluationPipelineOutput()
    data = epi.read_data(epi.data_path)
    model = epi.load_model(epi.model_input_path)
    if epi.keep_data:
        epo.data = data
    processed_data = epi.pre_process(data, model)
    if epi.keep_processed_data:
        epo.processed_data = processed_data
    evaluation_output = epi.evaluate(processed_data)
    if epi.keep_evaluation:
        epo.evaluation_output = evaluation_output
    return epo


def execute_pipeline(pi: PipelineInfo) -> PipelineOutput:
    mpi = pi.model_pipeline_info
    epi = pi.evaluation_pipeline_info
    mpo = execute_model_pipeline(mpi)
    epo = execute_evaluation_pipeline(epi)
    return PipelineOutput(model_pipeline_output=mpo, evaluation_pipeline_output=epo)



import pandas as pd
import pickle
from elizabeth.data_class import ModelPipelineInfo, EvaluationPipelineInfo
from elizabeth.executor import execute_model_pipeline, execute_evaluation_pipeline
from .setting import FIXTURE_DATA_PATH, FIXTURE_MODEL_PATH, TEST_MODEL_OUTPUT_PATH


class FakeModel:
    @staticmethod
    def predict(x_1, x_2):
        return 1 * x_1 + 2 * x_2


def fake_read_data(path):
    data = pd.read_csv(path)
    return data


def fake_pre_process(data):
    return data


def fake_training(_):
    return FakeModel


def fake_save_model(model, data_path):
    with open(data_path, 'wb') as f:
        pickle.dump(model, f)


def test_execute_model_pipeline():
    mpi = ModelPipelineInfo(model_output_path=TEST_MODEL_OUTPUT_PATH,
                            data_path=FIXTURE_DATA_PATH,
                            read_data=fake_read_data,
                            pre_process=fake_pre_process,
                            train=fake_training,
                            save_model=fake_save_model,
                            keep_data=True,
                            keep_processed_data=True,
                            keep_model=True)
    actual_mpo = execute_model_pipeline(mpi)
    expected_data = fake_read_data(FIXTURE_DATA_PATH)
    expected_processed_data = expected_data
    expected_model = FakeModel
    assert actual_mpo.data.equals(expected_data)
    assert actual_mpo.processed_data.equals(expected_processed_data)
    assert actual_mpo.model == expected_model


def fake_load_model(data_path):
    with open(data_path, 'rb') as f:
        model = pickle.load(f)
    return model


def fake_evaluation_pre_process(data, model):
    return data


def fake_evaluate(_):
    return {'acc': 0.9, 'recall': 0.8}


def test_execute_evaluation_pipeline():
    epi = EvaluationPipelineInfo(model_input_path=FIXTURE_MODEL_PATH,
                                 data_path=FIXTURE_DATA_PATH,
                                 read_data=fake_read_data,
                                 load_model=fake_load_model,
                                 pre_process=fake_evaluation_pre_process,
                                 evaluate=fake_evaluate)
    epo = execute_evaluation_pipeline(epi)
    expected_data = fake_read_data(FIXTURE_DATA_PATH)
    expected_processed_data = expected_data
    expected_evaluation_output = {'acc': 0.9, 'recall': 0.8}
    assert epo.data.equals(expected_data)
    assert epo.processed_data.equals(expected_processed_data)
    assert epo.evaluation_output == expected_evaluation_output



from setuptools import setup

setup(
    name="elizabeth",
    version="0.0.1",
    author="Shuhei Kishi",
    author_email="marubontan@gmail.com",
    description=("Machine leaning pipeline"),
    license="MIT",
    url="https://github.com/marubontan/elizabeth",
    packages=['elizabeth']
)
